/**********************************************************************
 Program Name:  FilterINTER.c
 Purpose:       Translates USENET/INTERNET format to Compuserve message
                  format for use with QL BBS Message Reader
 Author:        James Hunkins     Copyright 1994/1995
                                  JDH Technologies
                                  Freeware
 Input:         Name of text file containing USENET/INTERNET messages
                Name of output file for translating messages
 Output:        Message file matching Compuserve format for QL BBS
                  Message Reader
 EXEC Format:   EXEC FilterINTER
                EXEC FilterINTER;"original_mssg_file_name"
                EXEC FilterINTER;"original_mssg_file_name, translated_file_name"
                EXEC FilterINTER;"-F1 -T1 original_mssg_file_name"
                notes: .all control parameters in string are optional
                       . if translated_file_name is given, then original_file_name
                         must also be given
                       .-F, -T parameters must be before names
                       .-F, -T are case sensitive
                       .anything other than -F, -T are assumed to be file names
 Document:      FilterINTER_txt
 Update Info:   V0.0 08-22-94  Original
                               - pointer aware
                               - compiled with C68  V4.20
                V0.1 01-17-95  Modified to allow out-of-order header lines
                               - changed some skip message errors to missing info
                                 warnings, added for more header lines
                                 . line count missing still skips message

 Special Notes: . message numbers used (Message-ID) are alphanumeric
                  - QL Message reader expects numeric only
                  - assigns numeric numbers automatically and stores
                    translation in 'OutputFileName_mnt'
                  -used_mssg_no=5_chars_left_justify 5_spaces actual_message_no
                . FROM line can be long and complex so different forms available
                    from_style => default = 0
                      0 = full       jodyc#main1.main2.gen1.gen2 (Jody S. Anyone)
                      1 = short      jodyc
                      2 = address    jodyc#main1.main2.gen1.gen2
                      3 = real name                             Jody S. Anyone
                . normal TO line not available, so options for alternatives
                    to_style => default = 0
                      0 = Followup-To line
                      1 = Organization line
                . starts reading header at start of file or after reading mssg
                    header lines start with keyword:
                    header ends with blank line

 Progress Note:

 **********************************************************************/

#include <stdio.h>
#include <stdlib_h>
#include <string_h>
#include <ctype_h>
#include <qdos_h>

/* the following strings are what key lines start with
   the _len term is the number of characters */
#define section_line      "Newsgroups:"     /* SECTION name */
#define section_line_len  11
#define new_mssg          "Path:"           /* START of a new message */
#define new_mssg_len      5
#define mssg_id           "Message-ID:"     /* message number */
#define mssg_id_len       11
#define follow_line       "Followup-To:"    /* Followup-To line */
#define follow_line_len   12
#define subj_line         "Subject:"        /* SUBJECT name */
#define subj_line_len     8
#define org_line          "Organization:"   /* Organization line */
#define org_line_len      13
#define from_line         "From:"           /* FROM name */
#define from_line_len     5
#define date_line         "Date:"           /* DATE */
#define date_line_len     5
#define mssg_len          "Lines:"          /* number of lines in message */
#define mssg_len_len      6
#define EOL               10                /* end of line character */
#define sect_limit        128               /* maximum number of section names */

#define false             0
#define true              1



/* no TO in these files - not used */
#define to_line           "To:"             /* to line */
#define to_line_len       3


/* C68 declarations */
char _prog_name[] = "INTERNET/USENET Filter";
char _copyright[] = "(c)1994 JD Hunkins";
char _version[] = "JDH Technologies  V0.10";
/* void consetup_title(chanid_t); */
void (*_consetup)() = consetup_title;
/* char *_endmsg = NULL; */

/* Global Variables */
FILE *in;                 /* input file to be translated */
FILE *out;                /* translated file */
FILE *fmnum;              /* message number translations */
  /* format:   used_mssg_no=5_chars_left_justify 5_spaces actual_message_no */

long QL_stdout  =  0;     /* QL channel for standard output */
long line_ctr = 0;        /* current line read in, used for error reporting */

int from_style = 0;   /*  0 = full     jodyc#main1.main2.gen1.gen2 (Jody S. Anyone)
                          1 = short    jodyc
                          2 = address  jodyc#main1.main2.gen1.gen2
                          3 = real name                             Jody S. Anyone  */

int to_style = 0;     /*  0 = Followup-To line
                          1 = Organization line   */
int  section_ptr = 0;     /* next entry in section array */
int  line_ct;             /* no. of lines in message */
int  mssg_count = 0;  /* counts how messages found */

/* the following are made global to avoid taking up stack space */
char line[255];
char file_name[101];
char string1[41], string2[41], string3[41];
char path[255];           /* holds PATH for output file */
char section_name[41];    /* holds SECTION name for output file */
char section[sect_limit][41];  /* pointer to up to section names found */
char section_number[41];  /* holds SECTION number for output file */
char subj[41];            /* holds SUBJECT name for output file */
char to[41];              /* holds TO name for output file */
char from[41];            /* holds FROM name for output file */
char date[41];            /* holds DATE for output file */
char mssg_no[81];         /* holds BBS format MESSAGE NUMBER for output file */


/* prototype statements */
int  translate();
void error_out(int finish, int err_no, char other[]);
void key_pause(int sound1);
int  strip(char line[]);

void process_path();
void process_from();
void process_section();
void process_subject();
void process_follow();
void process_date();
void process_org();
void process_length();
void process_id();



/***********************************************************************
 Function:  main
 Input:     argv[0] = program name
            argv[1] = input message file name (optional)
            argv[2] = output message file name (optional)
            argc = number of arguments given in EXEC line
 ***********************************************************************/

int main(int argc, char *argv[])
{
int  first_parameter = 1; /* first possible file name location */

  /* initialize and open files */
  QL_stdout = fgetchid(stdout);   /* get standard output channel no */

  if (argc > 1)   /* check for -F and -T parameters */
    {
    if (strncmp("-F", argv[1], 2) == 0)
      {
      if (isdigit(argv[1][2]))
        from_style = atoi(&(argv[1]+2));
      else
        error_out(1,5,argv[1]);
      if ((from_style > 3) || (from_style < 0))   /* -F value out of range */
        error_out(1,5,argv[1]);
      first_parameter = 2;
      }
    else if (strncmp("-T", argv[1], 2) == 0)
      {
      if (isdigit(argv[1][2]))
        to_style = atoi(&(argv[1]+2));
      else
        error_out(1,5,argv[1]);
      if ((to_style > 1) || (to_style < 0))   /* -T value out of range */
        error_out(1,5,argv[1]);
      first_parameter = 2;
      }
    if (argc > 2)
      {
      if (strncmp("-F", argv[2], 2) == 0)
        {
        if (isdigit(argv[2][2]))
          from_style = atoi(&(argv[2]+2));
        else
          error_out(1,5,argv[2]);
        if ((from_style > 3) || (from_style < 0))   /* -F value out of range */
          error_out(1,5,argv[2]);
        first_parameter = 3;
        }
      else if (strncmp("-T", argv[2], 2) == 0)
        {
        if (isdigit(argv[2][2]))
          to_style = atoi(&(argv[2]+2));
        else
          error_out(1,5,argv[2]);
        if ((to_style > 1) || (to_style < 0))   /* -T value out of range */
          error_out(1,5,argv[2]);
        first_parameter = 3;
        }
      }
    }

  if (argc - first_parameter > 2)   /* too many arguments */
    error_out(1,1,"");

  /* get input file name */
  if ((argc - first_parameter) == 0)
    {
    printf("Enter file containing messages: ");
    gets(line);
    sscanf(line, "%s", &file_name);
    printf("\n");
    }
  else
    strcpy(file_name, argv[first_parameter]);

  if (file_name[0] == 0)
    error_out(1,6,file_name);
  if ((in = fopen(file_name, "rt")) == NULL)
    error_out(1,2,file_name);

  /* get output file name */
  if ((argc - first_parameter) < 2)
    {
    printf("Enter destination file name [%s_mrd] : ", file_name);
    gets(line);

    if (line[0] == 0)
      strcat(file_name, "_mrd");
    else
      sscanf(line, "%s", &file_name);
    printf("\n");
    }
  else
    strcpy(file_name, argv[first_parameter+1]);
  if ((out = fopen(file_name, "wt")) == NULL)
    error_out(1,3,file_name);

  /* open message number translation file, output header line */
  strcpy(line, file_name);
  strcat(file_name, "_mnt");   /* add _mnt to file output name */
  if ((fmnum = fopen(file_name, "wt")) == NULL)
    error_out(1,3,file_name);
  fprintf(fmnum, "Message Number Translations: %s\n", line);
  fprintf(fmnum, "=======================================================\n");
  fprintf(fmnum, "Used No   Original No\n");

  /* translate file */
  if (translate() == 0)    /* calls translate function */
    {
    printf("No messages found");
    io_delete("file_name");   /* delete output file */
    }
  fclose(in);
  error_out(1,0,"\n\n> Program Closed <\n ");
}


/*********************************************************************
 Function:    translate
 Purpose:     reads, translates from INTERNET/USENET format to native
                format (CIS) for QL BBS Message reader
 Return:      number of messages translated
**********************************************************************/

int translate()
{
int  done = 0;            /* 0 = indicates not done */
int  found_blank;         /* 1 = blank line found at end of header */
int  i;

char *colon_pos;          /* position of first ':' in line */
int  header_found = false;
int  header_error;
char letter;


  /* read in messages and translate header/end format to native format (CIS) */

  printf("TRANSLATING MSSG #:    0");   /* message # and back spaces */


  while((done == 0) && (!feof(in)))  /* loop until message found or end of file */
    {
    /* initialize next message header info */
    to[0] = 0;
    date[0] = 0;
    mssg_no[0] = 0;
    subj[0] = 0;
    from[0] = 0;
    section_name[0] = 0;
    line_ct = -1;

    /* find start of message */

    /* check for first line with key word and process */
    /*   first check for ':' in first 13 characters
         - then first character in word
         - then actual word and handle            */
    while (!header_found)
      {
      if (feof(in))                 /* end of file, quit translate */
          break;

      fgets(line, 255, in);         /* read in next line */
      line_ctr = line_ctr + 1;      /* increment line counter */
      if ((colon_pos = strchr(line, ':')) == NULL)  /* no keyword */
        continue;
      if ((colon_pos - line) > 13)  /* no keyword */
        continue;
      strip(line);                  /* remove EOL type characters */

      letter = toupper(line[0]);
      switch ( letter )
        {
        case 'P' :                  /* Path: */
          if (strncmp(new_mssg, line, new_mssg_len) == 0)
            {
            header_found = true;
            process_path();
            }
          break;

        case 'F' :                  /* From: or Followup-To: */
          if (strncmp(from_line, line, from_line_len) == 0)
            {
            header_found = true;
            process_from();
            }
          else if (strncmp(follow_line, line, follow_line_len) == 0)
            {
            header_found = true;
            process_follow();
            }
          break;

        case 'N' :    /* Newsgroups: */
          if (strncmp(section_line, line, section_line_len) == 0)
            {
            header_found = true;
            process_section();
            }
          break;

        case 'S' :    /* Subjects: */
          if (strncmp(subj_line, line, subj_line_len) == 0)
            {
            header_found = true;
            process_subject();
            }
          break;

        case 'D' :    /* Date: */
          if (strncmp(date_line, line, date_line_len) == 0)
            {
            header_found = true;
            process_date();
            }
          break;

        case 'O' :    /* Organization: */
          if (strncmp(org_line, line, org_line_len) == 0)
            {
            header_found = true;
            process_org();
            }
          break;

        case 'L' :    /* Lines: */
          if (strncmp(mssg_len, line, mssg_len_len) == 0)
            {
            header_found = true;
            process_length();
            }
          break;

        case 'M' :    /* Message-ID: */
          if (strncmp(mssg_id, line, mssg_id_len) == 0)
            {
            header_found = true;
            process_id();
            }
          break;
        } /* switch */
      }   /* while */

    if (!header_found)  /* new header not found, exit translate */
      {
      done = 1;
      break;
      }
    header_found = false;  /* prepare for next header search */


    /* header has been started, scan rest of header until blank line */
    while (true)
      {
      if (feof(in))  /* end of file, quit translate */
          break;

      fgets(line, 255, in);       /* read in next line */
      line_ctr = line_ctr + 1;    /* increment line counter */

      /* check for blank line (end of header) */
      if (strip(line) == 0)
        {
        found_blank = 1;
        break;
        }

      /* first check for ':' in first 13 characters
         - then first character in word
         - then actual word and handle              */
      if ((colon_pos = strchr(line, ':')) == NULL)  /* no keyword */
        continue;
      if ((colon_pos - line) > 13)  /* no keyword */
        continue;

      letter = toupper(line[0]);
      switch ( letter )
        {
        case 'P' :    /* Path: */
          if (strncmp(new_mssg, line, new_mssg_len) == 0)
            process_path();
          break;

        case 'F' :    /* From: or Followup-To: */
          if (strncmp(from_line, line, from_line_len) == 0)
            process_from();
          else if (strncmp(follow_line, line, follow_line_len) == 0)
            process_follow();
          break;

        case 'N' :    /* Newsgroups: */
          if (strncmp(section_line, line, section_line_len) == 0)
            process_section();
          break;

        case 'S' :    /* Subjects: */
          if (strncmp(subj_line, line, subj_line_len) == 0)
            process_subject();
          break;

        case 'D' :    /* Date: */
          if (strncmp(date_line, line, date_line_len) == 0)
            process_date();
          break;

        case 'O' :    /* Organization: */
          if (strncmp(org_line, line, org_line_len) == 0)
            process_org();
          break;

        case 'L' :    /* Lines: */
          if (strncmp(mssg_len, line, mssg_len_len) == 0)
            process_length();
          break;

        case 'M' :    /* Message-ID: */
          if (strncmp(mssg_id, line, mssg_id_len) == 0)
            process_id();
          break;
        } /* switch */
      }   /* while */

    if (feof(in))
      {
      done = 1;
      break;
      }

    mssg_count = mssg_count + 1;    /* message found, so increment count */
    printf("\rTRANSLATING MSSG #: %4i", mssg_count);

    /* output to header to output file in native format */
    /* remove all CR and EOL characters first */

    strip(section_number);
    strip(section_name);
    strip(subj);
    strip(from);
    strip(to);
    strip(date);
    strip(mssg_no);

    /* check for sufficient header information */
    header_error = false;

    if (mssg_no[0] == 0)
      {
      strcpy(mssg_no, "999");
      error_out(0, -7, mssg_no);
      }

    if (from[0] == 0)
      {
      error_out(0, -1, mssg_no);
      strcpy(from, "Unknown");
      }

    if (to[0] == 0)
      {
      error_out(0, -2, mssg_no);
      strcpy(to, "Unknown");
      }

    if (subj[0] == 0)
      {
      error_out(0, -3, mssg_no);
      strcpy(subj, "Unknown");
      }

    if (section_name[0] == 0)
      {
      error_out(0, -4, mssg_no);
      strcpy(section_name, "Unknown");
      itoa(999, section_number);
      }

    if (date[0] == 0)
      {
      error_out(0, -8, mssg_no);
      strcpy(date, "xx-xxx-xxxx");
      }

    if (line_ct == -1)
      {
      error_out(0, -6, mssg_no);
      header_error = true;
      }


    if (!header_error)
      {
      fprintf(out, "#: %s %s/%s\n", mssg_no, section_number, section_name);
      fprintf(out, "    %s\n", date);
      fprintf(out, "Sb: %s\n", subj);
      fprintf(out, "Fm: %s\n", from);
      fprintf(out, "To: %s\n", to);
      }


    /* output message lines */
    if (header_error)                 /* skip over message */
      for (i = 0; i < line_ct; i++)
        fgets(line, 255, in);
    else                              /* output message */
      for (i = 0; i < line_ct; i++)
        {
        fgets(line, 255, in);
        if (!header_error)
          {
          strip(line);
          fprintf(out, "%s\n", line);
          }
        }

    } /* main while loop */

  return mssg_count;
}       /* end translate function */



/*********************************************************************
 Function:    process_path
 Purpose:     parses Path: line
 Input:       none
 Global:      line[] - line to be parsed
 Output:      none

 *********************************************************************/

void process_path()
{
  strcpy(path, &(line[new_mssg_len + 1]));
}



/*********************************************************************
 Function:    process_from
 Purpose:     parses From: line
 Input:       none
 Global:      line[] - line to be parsed
 Output:      none

 *********************************************************************/

void process_from()
{
char *pline;

  switch(from_style)
    {
    case 0: strncpy(from, &line[6], 40);  /* full line */
            break;
    case 1: strncpy(from, &line[6], 40);  /* short form */
            pline = strchr(from, '@');
            pline[0] = 0;  /* end of line at '@' character */
            break;
    case 2: strncpy(from, &line[6], 40);  /* address form */
            if ((pline = strchr(from, '(')) != 0)
              pline[-1] = 0;    /* terminate string 1 before '(' character */
            break;
    case 3: if ((pline = strchr(line, '(')) == 0)  /* real name */
              from[0] = 0;
            else
              {
              strncpy(from, pline+1, 40);
              if ((pline = strchr(from, ')')) != 0)
                pline[0] = 0;   /* remove closing ')' */
              }
            break;
    }
}



/*********************************************************************
 Function:    process_section
 Purpose:     parses Newsgroups: line
 Input:       none
 Global:      line[] - line to be parsed
 Output:      none

 *********************************************************************/

void process_section()
{
int i;

  strncpy(section_name, &line[12], 40);

  /* check for section name previously existing, if not store new one */
  if (section_ptr == 0)   /* first section name entry */
    {
    strcpy(section[section_ptr], section_name);
    itoa(0, section_number);
    section_ptr = ++section_ptr;
    }
  else
    {
    section_number[0] = 0;  /* indicates no match found */
    for (i = 0; i < section_ptr; i++)
      if (strcmp(section_name, section[i]) == 0)  /* section previously used */
        {
        itoa(i, section_number);
        break;
        }
    if (section_number[0] == 0)   /* no previous use of section name */
      if (section_ptr >= sect_limit)
        {
        error_out(0, -4, mssg_no);   /* too many section names, overflow */
        itoa(999, section_number);
        strcpy(section_name, "OVERFLOW");
        }
      else                        /* new section name, add and use */
        {
        itoa(section_ptr, section_number);
        strcpy(section[section_ptr++], section_name);
        }
     }
}



/*********************************************************************
 Function:    process_subject
 Purpose:     parses Subjects: line
 Input:       none
 Global:      line[] - line to be parsed
 Output:      none

 *********************************************************************/

void process_subject()
{
  strncpy(subj, &line[9], 40);
}



/*********************************************************************
 Function:    process_follow
 Purpose:     parses FollowUp-To: line
 Input:       none
 Global:      line[] - line to be parsed
 Output:      none
 Notes:       to_style 0 => this is TO line

 *********************************************************************/

void process_follow()
{
  if (to_style == 0)
    {
    line[13+40] = 0;  /* truncate just in case */
    strcpy(to, &(line[13]));
    }
}



/*********************************************************************
 Function:    process_date
 Purpose:     parses Date: line
 Input:       none
 Global:      line[] - line to be parsed
 Output:      none

 *********************************************************************/

void process_date()
{
char *pline;

  if ((pline = strchr(line, ',')) != 0)  /* includes day, skip day */
    sscanf(pline+1, "%40s %10s %10s", date, string1, string2);
  else
    sscanf(line+6, "%40s %10s %10s", date, string1, string2);

  strcat(date, "-");  /* add separators and merge date strings */
  strcat(date, string1);
  strcat(date, "-");
  strcat(date, string2);
}



/*********************************************************************
 Function:    process_org
 Purpose:     parses Organization: line
 Input:       none
 Global:      line[] - line to be parsed
 Output:      none
 Notes:       to_style 1 => this is TO line

 *********************************************************************/

void process_org()
{
  if (to_style == 1)
    strncpy(to, &line[14], 40);
}



/*********************************************************************
 Function:    process_length
 Purpose:     parses Lines: line
 Input:       none
 Global:      line[] - line to be parsed
 Output:      none

 *********************************************************************/

void process_length()
{
  sscanf(line, "%s %8i", string1, &line_ct);
}



/*********************************************************************
 Function:    process_id
 Purpose:     parses Message-ID: line
 Input:       none
 Global:      line[] - line to be parsed
 Output:      none
 Notes:       use mssg_count for mssg number, store actual mssg_no

 *********************************************************************/

void process_id()
{
  strncpy(mssg_no, &line[mssg_id_len+1], 80);
  fprintf(fmnum, "%5i     %s\n", mssg_count, mssg_no);
  itoa(mssg_count, mssg_no);
}




/*********************************************************************
 Function:    strip
 Purpose:     removes 0D and 0A characters from string
 Input:       line[] = normal string
 Output:      = 0 if new line is empty string

 *********************************************************************/

int strip(char line[])
{
char *pline;

    if ((pline = strchr(line, 0x0A)) != 0)
      pline[0] = 0;
    if ((pline = strchr(line, 0x0D)) != 0)
      pline[0] = 0;
    if (line[0] == 0)
      return 0;       /* empty string */
    return 1;
}


/*********************************************************************
 Function:    error_out
 Purpose:     outputs error messages
 Input:       finish < 0, return as normal, -(#) = current mssg
                     = 1, exit(1) from program
              err_no = 0, quiet exit from program
                     > 0, ERROR message
                     < 0, WARNING message

 *********************************************************************/

void error_out(int finish, int err_no, char other[])
{
int i;

  switch(err_no)
    {
    case 0 :
      printf("%s\n", other);
      break;
    case 1 :
      printf("\nERROR: too many arguments in command line...\n");
      break;
    case 2 :
      printf("\nERROR: could not open input file [%s]...\n", other);
      break;
    case 3 :
      printf("\nERROR: could not open output file [%s]...\n", other);
      break;
    case 4 :
      printf("\nERROR: unexpected End of File before message complete...\n");
      printf("   Skipping message [%s]  Line #%i\n", other, line_ctr);
      break;
    case 5 :
      printf("\nERROR: illegal command line parameter [%s]...\n", other);
      break;
    case 6 :
      printf("\nERROR: no file name given...\n", other);
      break;
    case -1 :
      printf("\nWARNING: FROM line in header is wrong/missing...\n");
      printf("   Message [%s]  About line #%i\n", other, line_ctr);
      key_pause(0);
      break;
    case -2 :
      printf("\nWARNING: TO line in header is wrong/missing\n");
      printf("   Message [%s]  About line #%i\n", other, line_ctr);
      key_pause(0);
      break;
    case -3 :
      printf("\nWARNING: SUBJECT line in header is wrong/missing...\n");
      printf("   Message [%s]  About line #%i\n", other, line_ctr);
      key_pause(0);
      break;
    case -4 :
      printf("\nWARNING: NEWSGROUP line in header is wrong/missing...\n");
      printf("   Message [%s]  About line #%i\n", other, line_ctr);
      key_pause(0);
      break;
    case -5 :
      printf("\nWARNING: Too many section names...\n");
      printf("  All messages after this will be section #999, 'OVERFLOW'\n");
      printf("    starting with message [%s]  About line #%i\n", other, line_ctr);
      key_pause(0);
      break;
    case -6 :
      printf("\nWARNING: Line count not found...\n");
      printf("   Skipping message [%s]  About line #%i\n", other, line_ctr);
      key_pause(0);
      break;
    case -7 :
      printf("\nWARNING: Message-ID not found...\n");
      printf("   Message [%s]  About line #%i\n", other, line_ctr);
      key_pause(0);
      break;
    case -8 :
      printf("\nWARNING: Date not found...\n");
      printf("   Message [%s]  About line #%i\n", other, line_ctr);
      key_pause(0);
      break;
    default :
      printf("\nERROR/WARNING: unknown error [line #%i]...\n", line_ctr);
      key_pause(0);
      break;
    }

  if (finish > 0)
    {
    if (err_no != 0)
      printf("  >> Program Terminated << \n\n");
    for (i=1; i<1000; i++);
    exit(1);
    }

}


/***********************************************
 key_press  : waits for key input
   - input  : sound1 = 1 -> beep, 0 -> no beep
   - output : none

************************************************/

void key_pause(int sound1)
{
  if (sound1 == 1)
    beep(2400,10000);
  printf("\nPress any key to continue...\n");
  getch();
  printf("\n\n");
}
